
class BaseCoder(object):
    derived_context = {}

    def encode(self, data):
        raise NotImplementedError

    def decode(self, data):
        raise NotImplementedError


class ConstantCoder(BaseCoder):
    def __init__(self, encoded='', decoded={}):
        self.encoded = encoded
        self.decoded = decoded

    def encode(self, data):
        return self.encoded

    def decode(self, data):
        return self.decoded

BlankCoder = ConstantCoder

class PlainCoder(BaseCoder):
    def encode(self, data):
        return data

    def decode(self, data):
        return data




"""
Install 'pycrypto' package to use this module
"""

from Crypto.Cipher import AES

from . import BaseCryptor

def add_padding(data, block_size):
    data_len = len(data)
    pad_len = (block_size - data_len) % block_size
    if pad_len == 0:
        pad_len = block_size
    padding = chr(pad_len)
    return ''.join((data, padding * pad_len))

def strip_padding(data):
    padding = data[-1]
    pad_len = ord(padding)
    return data[:-pad_len]

def cut_key(key, key_size):
    while len(key) < key_size: # ...
        key += chr(0) * key_size
    return key[:key_size]


class AESCryptor(BaseCryptor):
    BLOCK_SIZE = 16

    def __init__(self, bits=128, iv=''): # iv is useless for temporary data in usual case
        if not bits in (128, 192, 256):
            raise ValueError(bits) # make one
        self.key_size = bits / 8
        self.iv = cut_key(iv, self.BLOCK_SIZE)

    def encrypt_stream(self, stream, secret_key):
        secret_key = cut_key(secret_key, self.key_size)
        cipher = AES.new(secret_key, AES.MODE_CBC, self.iv)
        padded = add_padding(stream, self.BLOCK_SIZE)
        encrypted = cipher.encrypt(padded)
        return encrypted

    def decrypt_stream(self, stream, secret_key):
        secret_key = cut_key(secret_key, self.key_size)
        cipher = AES.new(secret_key, AES.MODE_CBC, self.iv)
        decrypted = cipher.decrypt(stream)
        return strip_padding(decrypted)

    def encrypt_data(self, data, private_key, secret_key):
        secret_key = cut_key(secret_key, self.key_size)
        iv = cut_key(private_key, self.BLOCK_SIZE)
        cipher = AES.new(secret_key, AES.MODE_CBC, iv)
        padded = add_padding(data, self.BLOCK_SIZE)
        encrypted = cipher.encrypt(padded)
        return encrypted

    def decrypt_data(self, data, private_key, secret_key):
        secret_key = cut_key(secret_key, self.key_size)
        iv = cut_key(private_key, self.BLOCK_SIZE)
        cipher = AES.new(secret_key, AES.MODE_CBC, iv)
        decrypted = cipher.decrypt(data)
        return strip_padding(decrypted)





import time
import hashlib
from . import Verifier, DividedData
from ..exceptions import *

class AuthenticationFormatCorrupted(AuthenticationError):
    pass

class AuthenticationHashInvalid(AuthenticationError):
    pass

class AuthenticationExpired(AuthenticationError):
    pass

def timehash(private_key, public_key, hextime):
    m = hashlib.sha1()
    m.update(private_key)
    m.update(public_key)
    m.update(hextime)
    return m.hexdigest()

class TimeHashVerifier(Verifier):
    def __init__(self, limits=(300, 180), now=time.time):
        self.__pastlimit, self.__futurelimit = limits
        self.now = now

    def divide_verification_and_data(self, raw_data, secret_key):
        try:
            vals = raw_data.split('$', 2) # no rsplit - rightmost value can be bytes
        except ValueError:
            raise AuthenticationFormatCorrupted
        return DividedData(vals[0] + '$' + vals[1], vals[2])

    def public_key_from_verification(self, verification, secret_key):
        try:
            public_key, others = verification.split('$', 1)
        except ValueError:
            raise AuthenticationFormatCorrupted
        return public_key

    def verify(self, destructed, private_key, secret_key):
        public_key, others = destructed.verification.split('$', 1)
        timehex = others[:8]
        hexhash = others[8:]
        check_hexhash = timehash(private_key, public_key, timehex)
        if not hexhash == check_hexhash:
            raise AuthenticationHashInvalid
        given_time = int(timehex, 16)
        self.derived_context = {'time': given_time}
        now = self.now()
        if not -self.__pastlimit <= now - given_time <= self.__futurelimit:
            raise AuthenticationExpired(now=now, given=given_time, diff=now - given_time, past_limit=self.__pastlimit, future_limit=self.__futurelimit)
        return True

    def merge_verification_data(self, verification, raw_data, secret_key):
        return '$'.join((verification, raw_data))

    def encode_verification(self, private_key, public_key, secret_key):
        now = self.now()
        inow = int(now)
        hextime = '%8x' % inow
        hexhash = timehash(private_key, public_key, hextime)
        return str(''.join((public_key, '$', hextime, hexhash)))





from __future__ import absolute_import

import json
import collections
from . import BaseCoder

OrderedJsonDecoder = json.JSONDecoder(object_pairs_hook=collections.OrderedDict)

class JsonCoder(BaseCoder):
    def encode(self, data):
        return json.dumps(data, separators=(',',':')) # compact form

    def decode(self, data):
        return OrderedJsonDecoder.decode(data)



class BaseCryptor(object):
    derived_context = {}

    def encrypt_stream(self, stream, secret_key):
        raise NotImplementedError

    def decrypt_stream(self, stream, secret_key):
        raise NotImplementedError

    def encrypt_data(self, data, private_key, secret_key):
        raise NotImplementedError

    def decrypt_data(self, data, private_key, secret_key):
        raise NotImplementedError

    def decrypt_first_data(self, data, secret_key):
        return self.decrypt_data(data, secret_key, secret_key)

    def encrypt_first_data(self, data, secret_key):
        return self.encrypt_data(data, secret_key, secret_key)

class PlainCryptor(BaseCryptor):
    def encrypt_stream(self, stream, secret_key):
        return stream

    def decrypt_stream(self, stream, secret_key):
        return stream

    def encrypt_data(self, data, private_key, secret_key):
        return data

    def decrypt_data(self, data, private_key, secret_key):
        return data





from collections import namedtuple

DividedData = namedtuple('DividedData', ('verification', 'data'))
DestructedData = namedtuple('DestructedData', ('public_key', 'verification', 'data'))


class Verifier(object):
    derived_context = {}

    def divide_verification_and_data(self, raw_data, secret_key):
        raise NotImplementedError
        return DividedData(None, None) # expected interface

    def public_key_from_verification(self, verification, secret_key):
        raise NotImplementedError

    def destruct_data(self, raw_data, secret_key):
        vnd = self.divide_verification_and_data(raw_data, secret_key)
        public_key = self.public_key_from_verification(vnd.verification, secret_key)
        return DestructedData(public_key, vnd.verification, vnd.data)

    def destruct_first_data(self, raw_data, secret_key):
        return self.destruct_data(raw_data, secret_key)

    def verify(self, destructd, private_key, secret_key):
        raise NotImplementedError

    def encode_verification(self, private_key, public_key, secret_key):
        raise NotImplementedError

    def merge_verification_data(self, verification, raw_data, secret_key):
        raise NotImplementedError

    def construct_data(self, raw_data, private_key, public_key, secret_key):
        verification = self.encode_verification(private_key, public_key, secret_key)
        data = self.merge_verification_data(verification, raw_data, secret_key)
        return data

    def construct_first_data(self, raw_data, session_key, secret_key):
        return self.construct_data(raw_data, secret_key, session_key, secret_key)


class BypassVerifier(Verifier):
    def __init__(self, public_key='public_key'):
        self.public_key = public_key

    def divide_verification_and_data(self, raw_data, secret_key):
        return DividedData(None, raw_data)

    def public_key_from_verification(self, verification, secret_key):
        return self.public_key

    def verify(self, destructed, private_key, secret_key):
        return True

    def encode_verification(self, private_key, public_key, secret_key):
        return None

    def merge_verification_data(self, verification, raw_data, secret_key):
        return raw_data


class DataKeyVerifier(Verifier):
    """Supposed to used with PlainCoder"""
    def __init__(self, coder, key):
        self.coder = coder
        self.key = key

    def destruct_data(self, raw_data, secret_key):
        decoded_data = self.coder.decode(raw_data)
        verification = decoded_data[self.key]
        return DestructedData(verification, verification, decoded_data)

    def verify(self, verification, private_key, secret_key):
        return verification

    def construct_data(self, data, private_key, public_key, secret_key):
        data[self.key] = public_key
        return self.coder.encode(data)




import urlparse
from . import BaseCoder

class URLQueryCoder(BaseCoder):
    def encode(self, data):
        # No escape!
        return '&'.join('='.join(str(k), str(v)) for k, v in data.items())

    def decode(self, data):
        return urlparse.parse_qs(data, keep_blank_values=True)

class SimpleURLQueryCoder(BaseCoder):
    def encode(self, data):
        # No escape!
        return '&'.join('='.join((str(k), str(v))) for k, v in data.items())

    def decode(self, data):
        data = urlparse.parse_qs(data, keep_blank_values=True)
        for key in data:
            data[key] = data[key][0]
        return data



from .exceptions import *

class Authentication(object):
    def __init__(self, streamcoder, cryptor, verifier, datacoder, session_handler, secret_key):
        self.streamcoder = streamcoder
        self.cryptor = cryptor
        self.verifier = verifier
        self.datacoder = datacoder
        self.session_handler = session_handler
        self.secret_key = secret_key

    def _merge_context(self, context):
        for k, v in self.cryptor.derived_context.items():
            setattr(context, k, v)
        for k, v in self.verifier.derived_context.items():
            setattr(context, k, v)
        for k, v in self.datacoder.derived_context.items():
            setattr(context, k, v)

    def get_first_context(self, data, type=None, **params):
        """
        :param data: Stream data decodable by streamcoder
        :param type: Authentication type
        :return: Authentication context
        :rtype: :class:`instantauth.authenticate.Context`

        Decode process:

        | #0 Sender-Encoded Raw Stream |
                       |
                  streamcoder
                       V
        | #1 Sender-Encrypted Raw Data |
                       |
                    cryptor
                       V
        | #2 Receiver-Decrypted Raw Data | --verifier--> | Public Key |
                       |                                       |
                    verifier                                   V
                       V                                       |
          | #3 Sender-Encrypted Data | ---------->-------------+
                                                               |
                                                            cryptor
                                                               V
          | #5 Receiver-Decoded Data | <--coder-- | #4 Sender-Encoded Data |

        """
        #0: raw stream
        decoded_stream = self.streamcoder.decode(data) # 1:
        decrypted = self.cryptor.decrypt_stream(decoded_stream, self.secret_key) #2
        destructed = self.verifier.destruct_first_data(decrypted, self.secret_key) #3
        raw_data = self.cryptor.decrypt_first_data(destructed.data, self.secret_key) #4
        semantic_data = self.datacoder.decode(raw_data) #5
        context = Context(None, semantic_data)
        context.session = self.session_handler.session_from_session_key(destructed.public_key, type, **params)
        self._merge_context(context)
        return context

    def get_context(self, data):
        """
        :param data: Stream data decodable by streamcoder
        :return: Authentication context
        :rtype: :class:`instantauth.authenticate.Context`

        Decode process:

        | #0 Sender-Encoded Raw Stream |
                       |
                  streamcoder
                       V
        | #1 Sender-Encrypted Raw Data |
                       |
                    cryptor
                       V
        | #2 Receiver-Decrypted Raw Data | --verifier--> | Public Key | --session_handler--> | Session |
                       |                                       |                                  |
                    verifier                                   V                            session_handler
                       V                                       |                                  V
          | #3 Sender-Encrypted Data | ---------->-------------+--------------<------------ | Private Key |
                                                               |
                                                            cryptor
                                                               V
          | #5 Receiver-Decoded Data | <--coder-- | #4 Sender-Encoded Data |
        """
        #0: raw stream
        decoded_stream = self.streamcoder.decode(data) #1
        decrypted = self.cryptor.decrypt_stream(decoded_stream, self.secret_key) #2
        destructed = self.verifier.destruct_data(decrypted, self.secret_key) #3
        if not destructed.public_key:
            raise AuthenticationError
        try:
            session = self.session_handler.session_from_public_key(destructed.public_key)
        except KeyError:
            raise AuthenticationError('No public key found', key=destructed.public_key)
        if not session:
            raise AuthenticationError('No session found')

        private_key = self.session_handler.get_private_key(session)
        raw_data = self.cryptor.decrypt_data(destructed.data, private_key, self.secret_key) #4
        semantic_data = self.datacoder.decode(raw_data) #5
        if not self.verifier.verify(destructed, private_key, self.secret_key):
            raise AuthenticationError('Session verification failed')
        context = Context(session, semantic_data)
        self._merge_context(context)
        return context

    def build_first_data(self, data, session_key):
        """

        :param data:
        :param session_key:
        :rtype: :class:`bytes`

        Encode process:

        | #1 Original Data | --coder--> | #2 Encoded Data | --cryptor--> | #3 Encrypted Data |
                                                                                   |
                                                                                verifier
                                                                                   V
                           | #5 Encrypted Packed Data | <--cryptor-- | #4 Verification-Packed Data |
        """
        #1 data
        coded_data = self.datacoder.encode(data) #2
        encrypted_data = self.cryptor.encrypt_first_data(coded_data, self.secret_key) #3
        merged_data = self.verifier.construct_first_data(encrypted_data, session_key, self.secret_key) #4
        encrypted = self.cryptor.encrypt_stream(merged_data, self.secret_key) #5
        encoded_stream = self.streamcoder.encode(encrypted)
        return encoded_stream

    def build_data(self, data, session):
        """

        :param data: Structured data encodable by datacoder
        :param session: Authentication session
        :rtype: :class:`bytes`

        Encode process:

        | #1 Original Data | --coder--> | #2 Encoded Data | --cryptor--> | #3 Encrypted Data |
                                                                                   |
                                                                                verifier
                                                                                   V
                           | #5 Encrypted Packed Data | <--cryptor-- | #4 Verification-Packed Data |

        """
        #1: data
        private_key = self.session_handler.get_private_key(session)
        public_key = self.session_handler.get_public_key(session)
        coded_data = self.datacoder.encode(data) #2
        assert(coded_data is not None)
        encrypted_data = self.cryptor.encrypt_data(coded_data, private_key, self.secret_key) #3
        assert(encrypted_data is not None)
        merged_data = self.verifier.construct_data(encrypted_data, private_key, public_key, self.secret_key) #4
        encrypted = self.cryptor.encrypt_stream(merged_data, self.secret_key) #5
        encoded_stream = self.streamcoder.encode(encrypted)
        return encoded_stream


class Context(object):
    def __init__(self, session, data):
        self.session = session
        self.data = data

    def __repr__(self):
        return u'<Context({},{})>'.format(self.session, self.data)




class SessionHandler(object):
    def session_from_session_key(self, session_key, type=None):
        raise NotImplementedError

    def session_from_public_key(self, public_key):
        raise NotImplementedError

    def get_public_key(self, session):
        raise NotImplementedError

    def get_private_key(self, session):
        raise NotImplementedError



from __future__ import with_statement

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


def get_version():
    with open('instantauth/version.txt') as f:
        return f.read().strip()


def get_readme():
    try:
        with open('README.rst') as f:
            return f.read().strip()
    except IOError:
        return ''

setup(
    name='instantauth',
    version=get_version(),
    description='Not very credential authentication model.',
    long_description=get_readme(),
    author='Jeong YunWon',
    author_email='jeong+instantauth@youknowone.org',
    url='https://github.com/youknowone/instantauth',
    packages=(
        'instantauth',
        'instantauth/cryptors',
        'instantauth/verifiers',
        'instantauth/coders',
    ),
    package_data={
        'instantauth': ['version.txt']
    },
    install_requires=[
        'prettyexc',
    ],
)



from __future__ import absolute_import

import flask

from instantauth import Authentication

flask.Request.want_form_data_parsed = False

class FlaskAuthentication(Authentication):
    def from_request(self):
        raw_data = flask.request.data
        data = self.session_handler.decode_data(raw_data)
        context = self.get_context(data)
        flask.request.form = context.data
        flask.request.session = context.session
        return context

    def from_first_request(self, type=None, **params):
        raw_data = flask.request.data
        data = self.session_handler.decode_data(raw_data)
        context = self.get_first_context(data, type, **params)
        flask.request.form = context.data
        flask.request.session = context.session
        return context




from __future__ import absolute_import
import base64

from . import BaseCoder

class Base64Coder(BaseCoder):
    def encode(self, data):
        return base64.b64encode(data)

    def decode(self, data):
        return base64.b64decode(data)

